# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('hcore', '0010_timeseries_start_end_dates'),
    ]

    operations = [
        migrations.RenameField(
            model_name='timeseries',
            old_name='end_date',
            new_name='end_date_utc',
        ),
        migrations.RenameField(
            model_name='timeseries',
            old_name='start_date',
            new_name='start_date_utc',
        ),
    ]



import unittest
from django.contrib.auth.models import User, Group
from enhydris.hcore.models import Gentity

class PermissionsTestCase(unittest.TestCase):
    def setUp(self):
        self.object = Gentity.objects.create(name='testgent')
        self.object.save()
        self.user = User.objects.create(username='testuser')
        self.user.save()
        self.group = Group.objects.create(name='testgroup')
        self.group.save()

    def tearDown(self):
        self.object.delete()
        self.user.delete()
        self.group.delete()

    def testUserPerms(self):
        assert self.user.has_row_perm(self.object, 'permission') == False
        self.user.add_row_perm(self.object, 'permission')
        assert self.user.has_row_perm(self.object, 'permission') == True
        self.user.del_row_perm(self.object, 'permission')
        assert self.user.has_row_perm(self.object, 'permission') == False

    def testGroupPerms(self):
        assert self.user.has_row_perm(self.object, 'permission') == False
        assert self.group.has_row_perm(self.object, 'permission') == False
        self.group.user_set.add(self.user)
        self.group.add_row_perm(self.object, 'permission')
        assert self.group.has_row_perm(self.object, 'permission') == True
        assert self.user.has_row_perm(self.object, 'permission') == True
        self.group.del_row_perm(self.object, 'permission')
        assert self.user.has_row_perm(self.object, 'permission') == False
        assert self.group.has_row_perm(self.object, 'permission') == False






from django.contrib import admin
from enhydris.permissions.models import Permission

class PermissionAdmin(admin.ModelAdmin):
    model = Permission
    list_display = ('content_type', 'user', 'group', 'name')
    list_filter = ('name',)
    search_fields = ['object_id', 'content_type', 'user', 'group']
    raw_id_fields = ['user', 'group']

    def __unicode__(self):
        return u"%s | %s | %d | %s" % (self.content_type.app_label, self.content_type, self.object_id, self.name)

admin.site.register(Permission, PermissionAdmin)




# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('enhydris_synoptic', '0004_auto_20151104_1345'),
    ]

    operations = [
        migrations.RenameField(
            model_name='synopticgroup',
            old_name='name',
            new_name='slug',
        ),
        migrations.AlterField(
            model_name='synopticgroup',
            name='slug',
            field=models.SlugField(unique=True, help_text='Identifier to be used in URL'),
        ),
        migrations.AddField(
            model_name='synopticgroup',
            name='name',
            field=models.CharField(default='needs migration', max_length=50),
            preserve_default=False,
        ),
    ]



from datetime import datetime
import json

from django.contrib.auth.models import User, Permission
from django.contrib.contenttypes.models import ContentType
from django.test import TestCase
from django.test.client import MULTIPART_CONTENT, BOUNDARY, encode_multipart
from django.test.utils import override_settings

import iso8601
import pytz
from rest_framework.test import APITestCase

import enhydris
from enhydris.hcore import models
from enhydris.hcore.tests.test_views import RandomEnhydrisTimeseriesDataDir


def create_test_data():
    "Besides creating test data, return the list of created gentities."
    user1 = User.objects.create_user('user1', 'user1@nowhere.org', 'password1')
    user1.is_active = True
    user1.is_superuser = False
    user1.is_staff = False
    user1.save()

    user2 = User.objects.create_user('user2', 'user2@nowhere.org', 'password2')
    user2.is_active = True
    user2.is_superuser = False
    user2.is_staff = False
    user2.save()

    organization1 = models.Organization.objects.create(
        name="Test Organization")
    timezone1 = models.TimeZone.objects.create(code="Test Time Zone",
                                               utc_offset=2)
    variable1 = models.Variable.objects.create(descr="Test Variable")
    uom1 = models.UnitOfMeasurement.objects.create(
        descr="Test Unit Of Measurement", symbol="+")
    uom1.variables = [variable1]
    uom1.save()

    pd1 = models.PoliticalDivision.objects.create(
        name="Test Political Division")
    water_basin1 = models.WaterBasin.objects.create(name="Test Water Basin")
    water_division1 = models.WaterDivision.objects.create(
        name="Test Water Division")
    stype1 = models.StationType.objects.create(descr="Test Station Type")

    station1 = models.Station.objects.create(
        name="Test Station",
        water_division=water_division1,
        water_basin=water_basin1,
        political_division=pd1,
        owner=organization1,
        is_automatic=False,
        copyright_holder="Joe User",
        copyright_years=2014)
    station1.stype = [stype1]
    station1.save()

    station2 = models.Station.objects.create(
        name="Test Station 2",
        water_division=water_division1,
        water_basin=water_basin1,
        political_division=pd1,
        owner=organization1,
        is_automatic=False,
        copyright_holder="Joe User",
        copyright_years=2014
    )
    station2.stype = [stype1]
    station2.last_modified = iso8601.parse_date(
        "2010-05-10 14:26:22", default_timezone=pytz.utc)
    station2.save()

    # timeseries1
    models.Timeseries.objects.create(
        name="Test Timeseries",
        time_zone=timezone1,
        variable=variable1,
        unit_of_measurement=uom1,
        gentity=station1)
    # timeseries2
    models.Timeseries.objects.create(
        name="Test Timeseries 2",
        time_zone=timezone1,
        variable=variable1,
        unit_of_measurement=uom1,
        gentity=station1)

    # permissions1
    enhydris.permissions.models.Permission.objects.create(
        name="edit", object_id=station1.id, user=user1,
        content_type=ContentType.objects.get_for_model(models.Station))

    return [pd1, water_basin1, water_division1, station1, station2]


class ReadTestCase(APITestCase):

    def setUp(self):
        self.reference_gentities = create_test_data()

    def testGentityList(self):

        response = self.client.get("/api/Gentity/")
        response.data.sort(key=lambda x: x['id'])
        self.assertEqual(len(response.data), len(self.reference_gentities))
        for res_datum, ref_datum in zip(response.data,
                                        self.reference_gentities):
            self.assertEqual(res_datum['id'], ref_datum.id)
            self.assertEqual(res_datum['name'], ref_datum.name)
            self.assertEqual(res_datum['name_alt'], ref_datum.name_alt)
            self.assertEqual(res_datum['short_name'], ref_datum.short_name)
            self.assertEqual(res_datum['short_name_alt'],
                             ref_datum.short_name_alt)
            self.assertEqual(res_datum['remarks'], ref_datum.remarks)
            self.assertEqual(res_datum['remarks_alt'], ref_datum.remarks_alt)
            wb_id = (ref_datum.water_basin.id
                     if ref_datum.water_basin else None)
            self.assertEqual(res_datum['water_basin'], wb_id)
            wd_id = (ref_datum.water_division.id
                     if ref_datum.water_division else None)
            self.assertEqual(res_datum['water_division'], wd_id)
            pd_id = (ref_datum.political_division.id
                     if ref_datum.political_division else None)
            self.assertEqual(res_datum['political_division'], pd_id)
            res_last_modified = iso8601.parse_date(
                res_datum['last_modified'], default_timezone=pytz.utc)
            self.assertEqual(res_last_modified, ref_datum.last_modified)

        # Same thing, but test with modified_after
        n_all_gentities = len(self.reference_gentities)
        for item in self.reference_gentities[:]:
            if item.last_modified < datetime(2010, 5, 11, tzinfo=pytz.utc):
                self.reference_gentities.remove(item)
        n_recent_gentities = len(self.reference_gentities)
        self.assertTrue(n_recent_gentities < n_all_gentities)
        response = self.client.get(
            '/api/Gentity/modified_after/2010-05-11 00:00/')
        response.data.sort(key=lambda x: x['id'])
        self.assertEqual(len(response.data), len(self.reference_gentities))
        for res_datum, ref_datum in zip(response.data,
                                        self.reference_gentities):
            self.assertEqual(res_datum['id'], ref_datum.id)
            self.assertEqual(res_datum['name'], ref_datum.name)
            self.assertEqual(res_datum['name_alt'], ref_datum.name_alt)
            self.assertEqual(res_datum['short_name'], ref_datum.short_name)
            self.assertEqual(res_datum['short_name_alt'],
                             ref_datum.short_name_alt)
            self.assertEqual(res_datum['remarks'], ref_datum.remarks)
            self.assertEqual(res_datum['remarks_alt'], ref_datum.remarks_alt)
            wb_id = (ref_datum.water_basin.id
                     if ref_datum.water_basin else None)
            self.assertEqual(res_datum['water_basin'], wb_id)
            wd_id = (ref_datum.water_division.id
                     if ref_datum.water_division else None)
            self.assertEqual(res_datum['water_division'], wd_id)
            pd_id = (ref_datum.political_division.id
                     if ref_datum.political_division else None)
            self.assertEqual(res_datum['political_division'], pd_id)
            res_last_modified = iso8601.parse_date(res_datum['last_modified'],
                                                   default_timezone=None)
            self.assertEqual(res_last_modified, ref_datum.last_modified)


class WriteTestCase(TestCase):

    def setUp(self):
        create_test_data()
        self.timeseries1 = models.Timeseries.objects.get(
            name='Test Timeseries')

    def testTimeSeries(self):
        # Get an existing time series
        response = self.client.get(
            "/api/Timeseries/{}/".format(self.timeseries1.id))
        t = json.loads(response.content)

        # Change some of its attributes
        t['id'] = None
        t['name'] = "Test Timeseries 1221"
        t['remarks'] = "Yet another timeseries test"

        # Attempt to upload unauthenticated - should deny
        d = json.dumps(t)
        response = self.client.post("/api/Timeseries/", data=d,
                                    content_type='application/json')
        self.assertEqual(response.status_code, 403)
        self.assertEqual(models.Timeseries.objects.filter(
                         name='Test Timeseries 1221').count(), 0)

        # Now try again, this time logged on as user 2; again should deny
        self.assert_(self.client.login(username='user2', password='password2'))
        response = self.client.post("/api/Timeseries/", data=d,
                                    content_type='application/json')
        self.assertEqual(response.status_code, 403)
        self.assertEqual(models.Timeseries.objects.filter(
                         name='Test Timeseries 1221').count(), 0)
        self.client.logout()

        # Now try again, this time logged on as user 1; should accept
        self.assert_(self.client.login(username='user1', password='password1'))
        response = self.client.post("/api/Timeseries/", data=d,
                                    content_type='application/json')
        self.assertEqual(response.status_code, 201)
        self.assertEqual(models.Timeseries.objects.filter(
                         name='Test Timeseries 1221').count(), 1)
        self.client.logout()

    def testTimeseriesDelete(self):
        # Check the number of timeseries available
        ntimeseries = models.Timeseries.objects.count()

        # 1 is not enough; we need to know we aren't
        # deleting more than necessary.
        assert(ntimeseries > 1)

        # Attempt to delete unauthenticated - should fail
        response = self.client.delete(
            "/api/Timeseries/{}/".format(self.timeseries1.id))
        self.assertEqual(response.status_code, 403)
        self.assertEqual(models.Timeseries.objects.filter(
            pk=self.timeseries1.id).count(), 1)
        self.assertEqual(models.Timeseries.objects.count(), ntimeseries)

        # Try again as user2 - should fail
        self.assert_(self.client.login(username='user2', password='password2'))
        response = self.client.delete(
            "/api/Timeseries/{}/".format(self.timeseries1.id))
        self.assertEqual(response.status_code, 403)
        self.assertEqual(models.Timeseries.objects.filter(
            pk=self.timeseries1.id).count(), 1)
        self.assertEqual(models.Timeseries.objects.count(), ntimeseries)
        self.client.logout()

        # Try again as user1 - should succeed
        self.assert_(self.client.login(username='user1', password='password1'))
        response = self.client.delete(
            "/api/Timeseries/{}/".format(self.timeseries1.id))
        self.assertEqual(response.status_code, 204)
        self.assertEqual(models.Timeseries.objects.filter(
            pk=self.timeseries1.id).count(), 0)
        self.assertEqual(models.Timeseries.objects.count(), ntimeseries - 1)
        self.client.logout()

    @RandomEnhydrisTimeseriesDataDir()
    def testUploadTsDataUnauthenticated(self):
        # Attempt to upload some timeseries data, unauthenticated
        response = self.client.put(
            "/api/tsdata/{}/".format(self.timeseries1.id),
            encode_multipart(BOUNDARY,
                             {'timeseries_records': '2012-11-06 18:17,20,\n'}),
            content_type=MULTIPART_CONTENT)
        t = self.timeseries1.get_all_data()
        self.assertEqual(response.status_code, 403)
        self.assertEqual(len(t), 0)

    @RandomEnhydrisTimeseriesDataDir()
    def testUploadTsDataAsWrongUser(self):
        # Attempt to upload some timeseries data as user 2; should deny
        self.assert_(self.client.login(username='user2', password='password2'))
        response = self.client.put(
            "/api/tsdata/{}/".format(self.timeseries1.id),
            encode_multipart(BOUNDARY,
                             {'timeseries_records': '2012-11-06 18:17,20,\n'}),
            content_type=MULTIPART_CONTENT)
        t = self.timeseries1.get_all_data()
        self.assertEqual(response.status_code, 403)
        self.assertEqual(len(t), 0)
        self.client.logout()

    @RandomEnhydrisTimeseriesDataDir()
    def testUploadTsDataGarbage(self):
        self.assert_(self.client.login(username='user1', password='password1'))
        response = self.client.put(
            "/api/tsdata/{}/".format(self.timeseries1.id),
            encode_multipart(BOUNDARY,
                             {'timeseries_records': '2012-aa-06 18:17,20,\n'}),
            content_type=MULTIPART_CONTENT)
        t = self.timeseries1.get_all_data()
        self.assertEqual(response.status_code, 400)
        self.assertEqual(len(t), 0)
        self.client.logout()

    @RandomEnhydrisTimeseriesDataDir()
    def testUploadTsData(self):
        self.assert_(self.client.login(username='user1', password='password1'))
        response = self.client.put(
            "/api/tsdata/{}/".format(self.timeseries1.id),
            encode_multipart(BOUNDARY,
                             {'timeseries_records': '2012-11-06 18:17,20,\n'}),
            content_type=MULTIPART_CONTENT)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content, '1')
        t = models.Timeseries.objects.get(pk=self.timeseries1.id
                                          ).get_all_data()
        self.assertEqual(len(t), 1)
        self.assertEqual(t.items(0)[0], datetime(2012, 11, 06, 18, 17, 0))
        self.assertEqual(t.items(0)[1], 20)
        self.client.logout()

        # Append two more records
        self.assert_(self.client.login(username='user1', password='password1'))
        response = self.client.put(
            "/api/tsdata/{}/".format(self.timeseries1.id),
            encode_multipart(BOUNDARY,
                             {'timeseries_records':
                              '2012-11-06 18:18,21,\n2012-11-07 18:18,23,\n'}),
            content_type=MULTIPART_CONTENT)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content, '2')
        t = models.Timeseries.objects.get(pk=self.timeseries1.id
                                          ).get_all_data()
        self.assertEqual(len(t), 3)
        self.assertEqual(t.items(0)[0], datetime(2012, 11, 06, 18, 17, 0))
        self.assertEqual(t.items(0)[1], 20)
        self.assertEqual(t.items(1)[0], datetime(2012, 11, 06, 18, 18, 0))
        self.assertEqual(t.items(1)[1], 21)
        self.assertEqual(t.items(2)[0], datetime(2012, 11, 07, 18, 18, 0))
        self.assertEqual(t.items(2)[1], 23)
        self.client.logout()

        # Try to append an earlier record; should fail
        self.assert_(self.client.login(username='user1', password='password1'))
        response = self.client.put(
            "/api/tsdata/{}/".format(self.timeseries1.id),
            encode_multipart(BOUNDARY,
                             {'timeseries_records': '2012-11-05 18:18,21,\n'}),
            content_type=MULTIPART_CONTENT)
        self.client.logout()
        self.assertEqual(response.status_code, 400)
        t = models.Timeseries.objects.get(pk=self.timeseries1.id
                                          ).get_all_data()
        self.assertEqual(len(t), 3)
        self.client.logout()


class WriteStationTestCase(TestCase):

    def setUp(self):
        create_test_data()
        self.station1 = models.Station.objects.get(name="Test Station")

    def test_edit_station(self):
        # Get an existing station
        response = self.client.get("/api/Station/{}/".format(self.station1.id))
        station = json.loads(response.content)

        # Change some of its attributes
        station_id = station['id']
        del station['id']
        station['name'] = 'Test Station 1222'
        station['remarks'] = 'Yet another station test'

        # Attempt to upload unauthenticated - should deny
        d = json.dumps(station)
        response = self.client.put('/api/Station/{}/'.format(station_id),
                                   data=d, content_type='application/json')
        self.assertEqual(response.status_code, 403)
        self.assertEqual(models.Station.objects.filter(
                         name='Test Station 1222').count(), 0)

        # Try again, this time logged on as user 2; again should deny
        self.assertTrue(self.client.login(username='user2',
                                          password='password2'))
        response = self.client.put('/api/Station/{}/'.format(station_id),
                                   data=d, content_type='application/json')
        self.assertEqual(response.status_code, 403)
        self.assertEqual(models.Station.objects.filter(
                         name='Test Station 1222').count(), 0)
        self.client.logout()

        # Try again, as user 1; should accept
        self.assertTrue(self.client.login(username='user1',
                                          password='password1'))
        response = self.client.put('/api/Station/{}/'.format(station_id),
                                   data=d, content_type='application/json')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(models.Station.objects.filter(
                         name='Test Station 1222').count(), 1)
        self.client.logout()

    def test_station_delete(self):
        # Check the number of stations available
        nstations = models.Station.objects.count()

        # 1 is not enough; we need to know we aren't
        # deleting more than necessary.
        assert(nstations > 1)

        # Attempt to delete unauthenticated - should fail
        response = self.client.delete(
            '/api/Station/{}/'.format(self.station1.id))
        self.assertEqual(response.status_code, 403)
        self.assertEqual(models.Station.objects.filter(
            pk=self.station1.id).count(), 1)
        self.assertEqual(models.Station.objects.count(), nstations)

        # Try again as user2 - should fail
        self.assertTrue(self.client.login(username='user2',
                                          password='password2'))
        response = self.client.delete('/api/Station/{}/'.format(
            self.station1.id))
        self.assertEqual(response.status_code, 403)
        self.assertEqual(models.Station.objects.filter(
            pk=self.station1.id).count(), 1)
        self.assertEqual(models.Station.objects.count(), nstations)
        self.client.logout()

        # Try again as user1 - should succeed
        self.assertTrue(self.client.login(username='user1',
                                          password='password1'))
        response = self.client.delete(
            '/api/Station/{}/'.format(self.station1.id))
        self.assertEqual(response.status_code, 204)
        self.assertEqual(
            models.Station.objects.filter(pk=self.station1.id).count(), 0)
        self.assertEqual(models.Station.objects.count(), nstations - 1)
        self.client.logout()


class CreateStationTestCase(TestCase):

    def setUp(self):
        create_test_data()

        # Get an existing station
        obj = models.Station.objects.get(name='Test Station')
        response = self.client.get("/api/Station/{}/".format(obj.id))
        self.station = json.loads(response.content)

        # Change some of its attributes
        del self.station['id']
        self.station['name'] = 'Test Station 1507'
        self.station['remarks'] = 'Yet another station test'

    def test_create_unauthenticated(self):
        response = self.client.post('/api/Station/',
                                    data=json.dumps(self.station),
                                    content_type='application/json')
        self.assertEqual(response.status_code, 403)
        self.assertEqual(models.Station.objects.filter(
                         name='Test Station 1507').count(), 0)

    @override_settings(ENHYDRIS_USERS_CAN_ADD_CONTENT=True)
    def test_create_station_on_database_allowing_any_user(self):
        self.assertTrue(self.client.login(username='user2',
                                          password='password2'))
        response = self.client.post('/api/Station/',
                                    data=json.dumps(self.station),
                                    content_type='application/json')
        self.assertEqual(response.status_code, 201)
        self.assertEqual(models.Station.objects.filter(
            name='Test Station 1507').count(), 1)
        self.client.logout()

    @override_settings(ENHYDRIS_USERS_CAN_ADD_CONTENT=False)
    def test_create_station_by_disallowed_user(self):
        self.assertTrue(self.client.login(username='user2',
                                          password='password2'))
        response = self.client.post('/api/Station/',
                                    data=json.dumps(self.station),
                                    content_type='application/json')
        self.assertEqual(response.status_code, 403)
        self.assertEqual(models.Station.objects.filter(
            name='Test Station 1507').count(), 0)
        self.client.logout()

    @override_settings(ENHYDRIS_USERS_CAN_ADD_CONTENT=False)
    def test_create_station_by_allowed_user(self):
        # Give the appropriate permissions to the user
        user = User.objects.get(username='user2')
        permission = Permission.objects.get(codename='add_station')
        user.user_permissions.add(permission)
        user.save()

        self.assertTrue(self.client.login(username='user2',
                                          password='password2'))
        response = self.client.post('/api/Station/',
                                    data=json.dumps(self.station),
                                    content_type='application/json')
        self.assertEqual(response.status_code, 201)
        self.assertEqual(models.Station.objects.filter(
            name='Test Station 1507').count(), 1)
        self.client.logout()



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django import db
from django.db import migrations

from pthelma.timeseries import Timeseries


def copy_tsdata_from_db_to_datafile(apps, schema_editor):
    TimeseriesModel = apps.get_model('hcore', 'Timeseries')
    for timeseries in TimeseriesModel.objects.all():
        t = Timeseries(timeseries.id)
        t.read_from_db(db.connection)
        if len(t):
            timeseries.datafile.name = '{:010}'.format(timeseries.id)
            with open(timeseries.datafile.path, 'w') as f:
                t.write(f)
        else:
            timeseries.datafile = None
        timeseries.save()


def copy_tsdata_from_datafile_to_db(apps, schema_editor):
    TimeseriesModel = apps.get_model('hcore', 'Timeseries')
    for timeseries in TimeseriesModel.objects.all():
        t = Timeseries(timeseries.id)
        if not timeseries.datafile:
            continue
        with open(timeseries.datafile.path, 'r') as f:
            t.read(f)
        t.write_to_db(db.connection, transaction=db.transaction, commit=False)


class Migration(migrations.Migration):

    dependencies = [
        ('hcore', '0007_timeseries_datafile'),
    ]

    operations = [
        migrations.RunPython(copy_tsdata_from_db_to_datafile,
                             reverse_code=copy_tsdata_from_datafile_to_db),
    ]



"""
These template tags are for counting. Used in the frontpage stats
"""

from django.conf import settings
from django.template import Library, Node, TemplateSyntaxError
from django.db.models import get_model

from enhydris.hcore.models import *

register = Library()

def do_count(appmodel):
    try:
        model = get_model(*appmodel.split('.'))
        count = model.objects.all().count()
    except:
        count = 0

    return count

register.simple_tag( do_count )


class LastModifiedStations(Node):
    def __init__(self, number=5):
        self.number = number

    def render(self, context):
        try:
            station_objects = Station.objects.all()
            if len(settings.ENHYDRIS_SITE_STATION_FILTER)>0:
                station_objects = station_objects.filter(**settings.ENHYDRIS_SITE_STATION_FILTER)
            latest_stations = station_objects.all().order_by('last_modified').reverse()[:self.number]
        except ValueError:
            latest_stations = None

        context['latest_stations'] = latest_stations
        return ''

class DoGetLatestStations:

    def __init__(self):
        pass

    def __call__(self, parser, token):
        tokens = token.contents.split()
        if not tokens[1].isdigit():
            raise TemplateSyntaxError, (
                "The argument for '%s' must be an integer" % tokens[0])
        return LastModifiedStations(tokens[1])

register.tag('get_latest_stations', DoGetLatestStations())



from datetime import timedelta
import sys

from django.conf import settings
from django.contrib.auth.models import User, Group
from django.contrib.contenttypes.models import ContentType
from django.contrib.gis.db import models
from django.contrib.gis.db.backends import postgis
from django.core.files.storage import FileSystemStorage
from django.db.models import Q
from django.db.models import signals
from django.db.models.signals import post_save
from django.shortcuts import get_object_or_404
from django.utils.timezone import now
from django.utils.translation import ugettext_lazy as _
from django.utils._os import abspathu

import iso8601
from pthelma import timeseries
from pthelma.timeseries import IntervalType as it
from pthelma.timeseries import TimeStep as TTimeStep
from simpletail import ropen

from enhydris.permissions.models import Permission


#
# Lookups
#


class Lookup(models.Model):
    last_modified = models.DateTimeField(default=now, null=True,
                                         editable=False)
    descr = models.CharField(max_length=200, blank=True)
    descr_alt = models.CharField(max_length=200, blank=True)

    class Meta:
        abstract = True
        ordering = ('descr',)

    def __unicode__(self):
        return self.descr or self.descr_alt


#
# Lentity and descendants
#


def post_save_person_or_organization(sender, **kwargs):
    instance = kwargs['instance']
    try:
        string = instance.name
    except:
        string = instance.last_name + instance.first_name
    lentity = instance.lentity_ptr
    lentity.ordering_string = string
    super(Lentity, lentity).save()


class Lentity(models.Model):
    last_modified = models.DateTimeField(default=now, null=True,
                                         editable=False)
    remarks = models.TextField(blank=True)
    remarks_alt = models.TextField(blank=True)
    ordering_string = models.CharField(max_length=255, null=True, blank=True,
                                       editable=False)

    class Meta:
        verbose_name_plural = "Lentities"
        ordering = ('ordering_string',)

    def __unicode__(self):
        return (self.remarks or self.remarks_alt or self.name_any
                or str(self.id))

    @property
    def name_any(self):
        len_name = ""
        try:
            lentity = Person.objects.get(pk=self.id)
            len_name = lentity.last_name+" "+lentity.first_name
        except Person.DoesNotExist:
            try:
                lentity = Organization.objects.get(pk=self.id)
                len_name = lentity.name
            except Organization.DoesNotExist:
                pass

        return len_name


class Person(Lentity):
    last_name = models.CharField(max_length=100, blank=True)
    first_name = models.CharField(max_length=100, blank=True)
    middle_names = models.CharField(max_length=100, blank=True)
    initials = models.CharField(max_length=20, blank=True)
    last_name_alt = models.CharField(max_length=100, blank=True)
    first_name_alt = models.CharField(max_length=100, blank=True)
    middle_names_alt = models.CharField(max_length=100, blank=True)
    initials_alt = models.CharField(max_length=20, blank=True)
    f_dependencies = ['Lentity']

    class Meta:
        ordering = ('last_name', 'first_name',)

    def __unicode__(self):
        return self.last_name + ' ' + self.initials


post_save.connect(post_save_person_or_organization, sender=Person)


class Organization(Lentity):
    name = models.CharField(max_length=200, blank=True)
    acronym = models.CharField(max_length=50, blank=True)
    name_alt = models.CharField(max_length=200, blank=True)
    acronym_alt = models.CharField(max_length=50, blank=True)
    f_dependencies = ['Lentity']

    class Meta:
        ordering = ('name',)

    def __unicode__(self):
        return self.acronym if self.acronym else self.name


post_save.connect(post_save_person_or_organization, sender=Organization)


#
# Gentity and direct descendants
#


class Gentity(models.Model):
    last_modified = models.DateTimeField(default=now, null=True,
                                         editable=False)
    water_basin = models.ForeignKey('WaterBasin', null=True, blank=True)
    water_division = models.ForeignKey('WaterDivision', null=True, blank=True)
    political_division = models.ForeignKey('PoliticalDivision', null=True,
                                           blank=True)
    name = models.CharField(max_length=200, blank=True)
    short_name = models.CharField(max_length=50, blank=True)
    remarks = models.TextField(blank=True)
    name_alt = models.CharField(max_length=200, blank=True)
    short_name_alt = models.CharField(max_length=50, blank=True)
    remarks_alt = models.TextField(blank=True)

    class Meta:
        verbose_name_plural = "Gentities"
        ordering = ('name',)

    def __unicode__(self):
        return self.short_name or self.short_name_alt or self.name \
            or self.name_alt or str(self.id)


class Gpoint(Gentity):
    srid = models.IntegerField(null=True, blank=True)
    approximate = models.BooleanField(default=False)
    altitude = models.FloatField(null=True, blank=True)
    asrid = models.IntegerField(null=True, blank=True)
    f_dependencies = ['Gentity']
    point = models.PointField(null=True, blank=True)

    def save(self, force_insert=False, force_update=False, *args, **kwargs):
        super(Gpoint, self).save(force_insert, force_update, *args, **kwargs)

    def original_abscissa(self):
        if self.point:
            (x, y) = self.point.transform(self.srid, clone=True)
            return round(x, 2) if abs(x) > 180 and abs(y) > 90 else x
        else:
            return None

    def original_ordinate(self):
        if self.point:
            (x, y) = self.point.transform(self.srid, clone=True)
            return round(y, 2) if abs(x) > 180 and abs(y) > 90 else y
        else:
            return None

    def original_cs_name(self):
        if self.srid:
            srtext = postgis.models.PostGISSpatialRefSys.objects.get(
                srid=self.srid).srtext
            cstype, dummy, rest = srtext.partition('[')
            csname = rest.split('"')[1]
            return csname + ' (' + cstype + ')'


class Gline(Gentity):
    gpoint1 = models.ForeignKey(Gpoint, null=True, blank=True,
                                related_name='glines1')
    gpoint2 = models.ForeignKey(Gpoint, null=True, blank=True,
                                related_name='glines2')
    length = models.FloatField(null=True, blank=True)
    f_dependecies = ['Gentity']
    linestring = models.LineStringField(null=True, blank=True)


class Garea(Gentity):
    area = models.FloatField(null=True, blank=True)
    f_dependencies = ['Gentity']
    mpoly = models.MultiPolygonField(null=True, blank=True)

    def __unicode__(self):
        if self.area:
            return str(self.id) + " (" + str(self.area) + ")"
        return str(self.id)


#
# Gentity-related models
#


class PoliticalDivisionManager(models.Manager):
    """This is the default model Manager enhanced with utility methods."""

    def get_leaf_subdivisions(self, local_parent):
        """Gets a parent PoliticalDivision and returns all its tree leaves.

        Leaves are the last political division level elements. The method
        iterates on every node of the parent subtree and returns only the
        bottom level elements.
        CAUTION! This returns a "list" not a "QuerySet"!!!
        """

        child = None
        children = []
        object = PoliticalDivision.objects.get(pk=local_parent)
        parents = PoliticalDivision.objects.all().filter(
            Q(name=object.name) &
            Q(short_name=object.short_name) &
            Q(name_alt=object.name_alt) &
            Q(short_name_alt=object.short_name_alt))
        for parent in parents:
            children.append(parent)
            for child in self.filter(parent=parent.id):
                try:
                    children.extend(self.get_leaf_subdivisions(child.id))
                except TypeError:
                    # This is the case we have only an object instance,
                    # not a list
                    children.append(self.get_leaf_subdivisions(child.id))

        if children:
            return children
        if not child:
            return [p for p in parents]


class PoliticalDivision(Garea):
    parent = models.ForeignKey('self', null=True, blank=True)
    code = models.CharField(max_length=5, blank=True)

    objects = PoliticalDivisionManager()

    f_dependencies = ['Garea']

    def __unicode__(self):
        result = self.short_name or self.name or self.short_name_alt \
            or self.name_alt or str(self.id)
        if self.parent:
            result = result + ', ' + self.parent.__unicode__()
        return result


class WaterDivision(Garea):
    # TODO: Fill in the model
    f_dependencies = ['Garea']

    def __unicode__(self):
        return self.name or str(self.id)


class WaterBasin(Garea):
    parent = models.ForeignKey('self', null=True, blank=True)
    f_dependencies = ['Garea']

    def __unicode__(self):
        return self.name or str(self.id)


class GentityAltCodeType(Lookup):
    pass


class GentityAltCode(models.Model):
    last_modified = models.DateTimeField(default=now, null=True,
                                         editable=False)
    gentity = models.ForeignKey(Gentity)
    type = models.ForeignKey(GentityAltCodeType)
    value = models.CharField(max_length=100)

    def __unicode__(self):
        return self.type.descr+' '+self.value

    @property
    def related_station(self):
        try:
            return Station.objects.get(id=self.gentity.id)
        except:
            return None


class GentityGenericDataType(Lookup):
    file_extension = models.CharField(max_length=16)


class GentityGenericData(models.Model):
    last_modified = models.DateTimeField(default=now, null=True,
                                         editable=False)
    gentity = models.ForeignKey(Gentity)
    descr = models.CharField(max_length=100)
    remarks = models.TextField(blank=True)
    descr_alt = models.CharField(max_length=100)
    remarks_alt = models.TextField(blank=True)
    content = models.TextField(blank=True)
    data_type = models.ForeignKey(GentityGenericDataType)

    def __unicode__(self):
        return self.descr or self.descr_alt or str(self.id)

    @property
    def related_station(self):
        try:
            return Station.objects.get(id=self.gentity.id)
        except:
            return None


class FileType(Lookup):
    mime_type = models.CharField(max_length=64)

    def __unicode__(self):
        if self.mime_type:
            return self.mime_type
        return str(self.id)


class GentityFile(models.Model):
    last_modified = models.DateTimeField(default=now, null=True,
                                         editable=False)
    gentity = models.ForeignKey(Gentity)
    date = models.DateField(blank=True, null=True)
    file_type = models.ForeignKey(FileType)
    content = models.FileField(upload_to='gentityfile')
    descr = models.CharField(max_length=100)
    remarks = models.TextField(blank=True)
    descr_alt = models.CharField(max_length=100)
    remarks_alt = models.TextField(blank=True)

    def __unicode__(self):
        return self.descr or self.descr_alt or str(self.id)

    @property
    def related_station(self):
        try:
            return Station.objects.get(id=self.gentity.id)
        except:
            return None


class EventType(Lookup):
    pass


class GentityEvent(models.Model):
    last_modified = models.DateTimeField(default=now, null=True,
                                         editable=False)
    gentity = models.ForeignKey(Gentity)
    date = models.DateField()
    type = models.ForeignKey(EventType)
    user = models.CharField(max_length=64)
    report = models.TextField(blank=True)
    report_alt = models.TextField(blank=True)

    class Meta:
        ordering = ['-date']

    def __unicode__(self):
        return str(self.date)+' '+self.type.__unicode__()

    @property
    def related_station(self):
        try:
            return Station.objects.get(id=self.gentity.id)
        except:
            return None


#
# Station and its related models
#


class StationType(Lookup):
    pass


class StationManager(models.GeoManager):
    """Default manager enhanced with utility methods."""

    def get_by_political_division(self, political_division):
        """Get all stations which belong to the specific political division."""
        leaves = PoliticalDivision.objects.get_leaf_subdivisions(
            political_division)
        return Station.objects.filter(political_division__in=leaves)


def handle_maintainer_permissions(sender, instance, **kwargs):

    if settings.ENHYDRIS_USERS_CAN_ADD_CONTENT:
        old_perms = Permission.objects.filter(
            object_id=instance.pk,
            content_type=ContentType.objects.get_for_model(Station))

        new_users = instance.maintainers.all()
        # remove all old perms for maintainers. keep for creator
        for p in old_perms:
            if not p.user == instance.creator:
                p.delete()

        # add new perms
        [u.add_row_perm(instance, 'edit') for u in new_users]


class Station(Gpoint):
    owner = models.ForeignKey(Lentity, related_name="owned_stations")
    stype = models.ManyToManyField(StationType, verbose_name='type')
    is_automatic = models.BooleanField(default=False)
    start_date = models.DateField(null=True, blank=True)
    end_date = models.DateField(null=True, blank=True)
    overseers = models.ManyToManyField(Person, through='Overseer',
                                       related_name='stations_overseen')
    copyright_holder = models.TextField()
    copyright_years = models.CharField(max_length=10)
    # The following two fields are only useful when USERS_CAN_ADD_CONTENT
    # is set.
    creator = models.ForeignKey(User, null=True, blank=True,
                                related_name='created_stations')
    maintainers = models.ManyToManyField(User, blank=True,
                                         related_name='maintaining_stations')

    f_dependencies = ['Gpoint']
    # Managers
    objects = StationManager()

    def show_overseers(self):
        return " ".join([i.__unicode__() for i in self.overseers.all()])

    show_overseers.short_description = "List of Overseers"


signals.post_save.connect(handle_maintainer_permissions, Station)


class Overseer(models.Model):
    last_modified = models.DateTimeField(default=now, null=True,
                                         editable=False)
    station = models.ForeignKey(Station)
    person = models.ForeignKey(Person)
    is_current = models.BooleanField(default=False)
    start_date = models.DateField(null=True, blank=True)
    end_date = models.DateField(null=True, blank=True)

    def __unicode__(self):
        return self.person.__unicode__()


class InstrumentType(Lookup):
    pass


class Instrument(models.Model):

    class Meta:
        ordering = ('name',)

    last_modified = models.DateTimeField(default=now, null=True,
                                         editable=False)
    station = models.ForeignKey(Station)
    type = models.ForeignKey(InstrumentType)
    manufacturer = models.CharField(max_length=50, blank=True)
    model = models.CharField(max_length=50, blank=True)
    start_date = models.DateField(null=True, blank=True)
    end_date = models.DateField(null=True, blank=True)
    name = models.CharField(max_length=100, blank=True)
    remarks = models.TextField(blank=True)
    name_alt = models.CharField(max_length=100, blank=True)
    remarks_alt = models.TextField(blank=True)

    def __unicode__(self):
        # This gets commended out because it causes significant delays #XXX:
        # return self.name or self.name_alt or self.type.descr or str(self.id)
        return self.name or str(self.id)


#
# Time series and related models
#


class Variable(Lookup):
    pass


class UnitOfMeasurement(Lookup):
    symbol = models.CharField(max_length=50)
    variables = models.ManyToManyField(Variable)

    def __unicode__(self):
        if self.symbol:
            return self.symbol
        return str(self.id)


if sys.version_info >= (3, 0):
    from datetime import timezone
else:
    from datetime import tzinfo

    class timezone(tzinfo):

        def __init__(self, offset, name=None):
            self.offset = offset
            self.name = name

        def utcoffset(self, dt):
            return self.offset

        def dst(self, dt):
            return None

        def tzname(self, dt):
            return self.name

        def fromutc(self, dt):
            return dt + self.offset

        def __str__(self):
            return '{} ({:+03}{:02})'.format(
                self.name,
                int(self.offset.total_seconds() / 3600),
                int((self.offset.total_seconds() % 3600) / 60))

        def __repr__(self):
            return '<{}.{} object: {}>'.format(
                self.__class__.__module__, self.__class__.__name__,
                self.__str__())


class TimeZone(models.Model):
    last_modified = models.DateTimeField(default=now, null=True,
                                         editable=False)
    code = models.CharField(max_length=50)
    utc_offset = models.SmallIntegerField()

    @property
    def as_tzinfo(self):
        return timezone(timedelta(minutes=self.utc_offset), self.code)

    def __unicode__(self):
        return '%s (UTC%+03d%02d)' % (
            self.code,
            (abs(self.utc_offset) / 60) * (-1 if self.utc_offset < 0 else 1),
            abs(self.utc_offset % 60),)

    class Meta:
        ordering = ('utc_offset',)


def _int_xor(i1, i2):
    """Return True if one and only one of i1 and i2 is zero."""
    return (i1 or i2) and not (i1 and i2)


class TimeStep(Lookup):
    length_minutes = models.PositiveIntegerField()
    length_months = models.PositiveSmallIntegerField()

    def __unicode__(self):
        """
        Return timestep descriptions in a more human readable format
        """
        days = self.length_minutes / 1440
        hours = self.length_minutes / 60 - 24 * days
        minutes = self.length_minutes - 60 * hours - 1440 * days
        years = self.length_months / 12
        months = self.length_months - years*12
        if self.descr != "":
            desc = self.descr+" - "
        else:
            desc = ""
        link = ""
        if years:
            desc += "%d year(s)" % years
            link = ','
        if months:
            desc += link+"%d month(s)" % months
            link = ','
        if days:
            desc += link+"%d day(s)" % days
            link = ','
        if hours:
            desc += link+"%d hour(s)" % hours
            link = ','
        if minutes:
            desc += link+"%d minute(s)" % minutes
            link = ','
        if desc:
            return desc
        return "(0,0)"

    def save(self, force_insert=False, force_update=False, *args, **kwargs):
        if not _int_xor(self.length_minutes, self.length_months):
            raise Exception(_(
                "%s is not a valid time step; exactly one of minutes and "
                "months must be zero") % self.__unicode__())
        super(TimeStep, self).save(force_insert, force_update, *args, **kwargs)


class IntervalType(Lookup):
    value = models.CharField(max_length=50)

    def __unicode__(self):
        return self.descr


class TimeseriesStorage(FileSystemStorage):
    """Stores timeseries data files in settings.ENHYDRIS_TIMESERIES_DATA_DIR.

    TimeseriesStorage() is essentially the same as
    FileSystemStorage(location=settings.ENHYDRIS_TIMESERIES_DATA_DIR)), with
    the difference that it checks the value of ENHYDRIS_TIMESERIES_DATA_DIR
    each time it does something. This allows the setting to be overridden
    in unit tests with override_settings(). If using FileSystemStorage,
    it would not be overriddable, because "location" would be set when
    models.py is read and that would be it.
    """

    def path(self, name):
        self.location = abspathu(settings.ENHYDRIS_TIMESERIES_DATA_DIR)
        return super(TimeseriesStorage, self).path(name)


class Timeseries(models.Model):
    last_modified = models.DateTimeField(default=now, null=True,
                                         editable=False)
    gentity = models.ForeignKey(Gentity, related_name="timeseries")
    variable = models.ForeignKey(Variable)
    unit_of_measurement = models.ForeignKey(UnitOfMeasurement)
    name = models.CharField(max_length=200, blank=True)
    name_alt = models.CharField(max_length=200, blank=True, default='')
    hidden = models.BooleanField(null=False, blank=False, default=False)
    precision = models.SmallIntegerField(null=True, blank=True)
    time_zone = models.ForeignKey(TimeZone)
    remarks = models.TextField(blank=True)
    remarks_alt = models.TextField(blank=True, default='')
    instrument = models.ForeignKey(Instrument, null=True, blank=True)
    time_step = models.ForeignKey(TimeStep, null=True, blank=True)
    interval_type = models.ForeignKey(IntervalType, null=True, blank=True)
    timestamp_rounding_minutes = models.PositiveIntegerField(
        null=True, blank=True, help_text=_(
            "For an hourly time series whose timestamps end in :00, set this "
            "to zero; if they end in :12, set it to 12. For a ten-minute time "
            "series with timestamps ending in :12, :22, :32, etc., set it to "
            "2.  For daily ending at 08:00, set it to 480. Leave empty if "
            "timestamps are irregular."
        )
    )
    timestamp_rounding_months = models.PositiveSmallIntegerField(
        null=True, blank=True, help_text=_(
            "Set this to zero, except for annual time series, indicating the "
            "difference from January; for example, set it to 9 if the "
            "timestamps use a hydrological year starting in October. Leave "
            "empty if timestamps are irregular."
        )
    )
    timestamp_offset_minutes = models.IntegerField(
        null=True, blank=True, help_text=_(
            "If unsure, set this to zero. It indicates the difference of "
            "what is shown from what is meant. For example, if for an hourly "
            "time series it is -5, then 2015-10-14 11:00 means the interval "
            "from 2015-10-14 09:55 to 2015-10-14 10:55. -1440 is common for "
            "daily time series."
        )
    )
    timestamp_offset_months = models.SmallIntegerField(
        null=True, blank=True, help_text=_(
            "If unsure, set this to 1 for monthly, 12 for annual, and zero "
            "otherwise.  For a monthly time series, an offset of -475 "
            "minutes and 1 month means that 2003-11-01 00:00 (normally "
            "shown as 2003-11) denotes the interval 2003-10-31 18:05 to "
            "2003-11-30 18:05."
        )
    )
    datafile = models.FileField(null=True, blank=True,
                                storage=TimeseriesStorage())
    start_date_utc = models.DateTimeField(null=True, blank=True)
    end_date_utc = models.DateTimeField(null=True, blank=True)

    class Meta:
        verbose_name = "Time Series"
        verbose_name_plural = "Time Series"
        ordering = ('hidden',)

    @property
    def start_date(self):
        return self.start_date_utc.astimezone(tz=self.time_zone.as_tzinfo)

    @property
    def end_date(self):
        return self.end_date_utc.astimezone(tz=self.time_zone.as_tzinfo)

    def set_start_and_end_date(self):
        if (not self.datafile) or (self.datafile.size < 10):
            self.start_date_utc = None
            self.end_date_utc = None
            return
        with open(self.datafile.path, 'r') as f:
            self.start_date_utc = iso8601.parse_date(
                f.readline().split(',')[0],
                default_timezone=self.time_zone.as_tzinfo)
        with ropen(self.datafile.path, bufsize=80) as f:
            self.end_date_utc = iso8601.parse_date(
                f.readline().split(',')[0],
                default_timezone=self.time_zone.as_tzinfo)

    def get_empty_timeseries_object(self):
        sign = -1 if self.time_zone.utc_offset < 0 else 1
        try:
            location = {
                'abscissa': self.gentity.gpoint.point[0],
                'ordinate': self.gentity.gpoint.point[1],
                'srid': self.gentity.gpoint.srid,
                'altitude': self.gentity.gpoint.altitude,
                'asrid': self.gentity.gpoint.asrid,
            }
        except TypeError:
            # TypeError occurs when self.gentity.gpoint.point is None,
            # meaning the co-ordinates aren't registered.
            location = None
        t = timeseries.Timeseries(
            id=self.id,
            time_step=ReadTimeStep(self.id, self),
            unit=self.unit_of_measurement.symbol,
            title=self.name,
            timezone='{} (UTC{:+03d}{:02d})'.format(
                self.time_zone.code,
                abs(self.time_zone.utc_offset) / 60 * sign,
                abs(self.time_zone.utc_offset) % 60),
            variable=self.variable.descr,
            precision=self.precision,
            location=location,
            comment='%s\n\n%s' % (self.gentity.name, self.remarks))
        return t

    def get_all_data(self):
        t = self.get_empty_timeseries_object()
        if self.datafile:
            with open(self.datafile.path, 'r') as f:
                t.read(f)
        return t

    def set_data(self, data):
        t = self.get_empty_timeseries_object()
        t.read(data)
        if not self.datafile:
            self.datafile.name = '{:010}'.format(self.id)
        with open(self.datafile.path, 'w') as f:
            t.write(f)
        self.save()
        return len(t)

    def append_data(self, data):
        if not self.datafile:
            return self.set_data(data)
        t = self.get_empty_timeseries_object()
        t.read(data)
        if not len(t):
            return 0
        with ropen(self.datafile.path, bufsize=80) as f:
            old_data_end_date = iso8601.parse_date(f.readline().split(',')[0]
                                                   ).replace(tzinfo=None)
        new_data_start_date = t.bounding_dates()[0]
        if old_data_end_date >= new_data_start_date:
            raise ValueError((
                "Cannot append time series: "
                "its first record ({}) has a date earlier than the last "
                "record ({}) of the timeseries to append to.")
                .format(new_data_start_date, old_data_end_date))
        with open(self.datafile.path, 'a') as f:
            t.write(f)
        self.save()
        return len(t)

    def get_first_line(self):
        if not self.datafile or self.datafile.size < 10:
            return ''
        with open(self.datafile.path, 'r') as f:
            return f.readline()

    def get_last_line(self):
        if not self.datafile or self.datafile.size < 10:
            return ''
        with ropen(self.datafile.path, bufsize=80) as f:
            lastline = f.readline()
            return lastline if len(lastline) > 5 else f.readline()

    @property
    def related_station(self):
        try:
            return Station.objects.get(id=self.gentity.id)
        except:
            return None

    def __unicode__(self):
        return self.name

    def save(self, force_insert=False, force_update=False, *args, **kwargs):
        if not self.time_step:
            if self.timestamp_rounding_minutes \
                    or self.timestamp_rounding_months \
                    or self.timestamp_offset_minutes \
                    or self.timestamp_offset_months:
                raise Exception(_("Invalid time step: if time_step is null, "
                                  "rounding and offset must also be null"))
        else:
            if self.timestamp_offset_minutes is None \
                    or self.timestamp_offset_months is None:
                raise Exception(_("Invalid time step: if time_step is not "
                                  "null, offset must be provided"))
            if ((self.timestamp_rounding_minutes is None
                 and self.timestamp_rounding_months is not None)
                or (self.timestamp_rounding_minutes is not None
                    and self.timestamp_rounding_months is None)):
                raise Exception(_("Invalid time step: roundings must be "
                                  "both null or not null"))
        self.set_start_and_end_date()
        super(Timeseries, self).save(force_insert, force_update, *args,
                                     **kwargs)


# BlobField is obsolete, but migration 0009 can't work without it.

class BlobField(models.Field):

    def db_type(self, connection):
        return 'bytea'


# Profile creation upon user registration
def user_post_save(sender, instance, **kwargs):
    profile, new = UserProfile.objects.get_or_create(user=instance)


def make_user_editor(sender, instance, **kwargs):
    user = instance
    if not user.is_superuser:
        group, created = Group.objects.get_or_create(name='editors')
        user.groups.add(group)


def ReadTimeStep(id, timeseries_instance=None):
    """Read time step from Timeseries Object."""
    if timeseries_instance is None:
        timeseries_instance = get_object_or_404(Timeseries, pk=int(id))
    t = timeseries_instance
    return TTimeStep(
        length_minutes=t.time_step.length_minutes if t.time_step else 0,
        length_months=t.time_step.length_months if t.time_step else 0,
        timestamp_rounding=(None if None in (t.timestamp_rounding_minutes,
                                             t.timestamp_rounding_months)
                            else (t.timestamp_rounding_minutes,
                                  t.timestamp_rounding_months)),
        timestamp_offset=(None if None in (t.timestamp_offset_minutes,
                                           t.timestamp_offset_months)
                          else (t.timestamp_offset_minutes,
                                t.timestamp_offset_months)),
        interval_type=(None if not t.interval_type else
                       {'sum': it.SUM,
                        'average': it.AVERAGE,
                        'vector_average': it.VECTOR_AVERAGE,
                        'minimum': it.MINIMUM,
                        'maximum': it.MAXIMUM}[t.interval_type.value.lower()]))


class UserProfile(models.Model):
    user = models.OneToOneField(User, verbose_name=_('Username'))
    fname = models.CharField(_('First Name'), null=True, blank=True,
                             max_length=30)
    lname = models.CharField(_('Last Name'), null=True, blank=True,
                             max_length=30)
    address = models.CharField(_('Location'), null=True, blank=True,
                               max_length=100)
    organization = models.CharField(_('Organization'), null=True, blank=True,
                                    max_length=100)
    email_is_public = models.BooleanField(default=False)

    def __unicode__(self):
        name = self.user.get_full_name()
        if name:
            return unicode("%s" % self.user.get_full_name())
        else:
            return unicode("%s" % self.user.username)

    class Meta:
        verbose_name = _('Profile')
        verbose_name_plural = _('Profiles')

    def first_name(self):
        return u"%s" % self.fname

    def last_name(self):
        return u"%s" % self.lname

    def full_name(self):
        if (self.fname is None and self.lname is None) \
                or (self.fname == self.lname == ''):
            return "No name specified"
        return u"%s %s" % (self.fname, self.lname)

    def email(self):
        return u"%s" % self.user.email


signals.post_save.connect(user_post_save, User)

if settings.ENHYDRIS_USERS_CAN_ADD_CONTENT:
    signals.post_save.connect(make_user_editor, User)



from django import template
from django.utils.safestring import SafeUnicode

register=template.Library()

@register.filter
def in_group(user, group):
    """Returns True/False if the user is in the given group(s).
    Usage::
        {% if user|in_group:"Friends" %}
        or
        {% if user|in_group:"Friends,Enemies" %}
        ...
        {% endif %}
    You can specify a single group or comma-delimited list.
    No white space allowed.
    """
    import re
    if re.search(',', group): group_list = re.sub('\s+','',group).split(',')
    elif re.search(' ', group): group_list = group.split()
    else: group_list = [group]
    user_groups = []
    for group in user.groups.all(): user_groups.append(str(group.name))
    if filter(lambda x:x in user_groups, group_list): return True
    else: return False
in_group.is_safe = True




import os
import tempfile


ROOT_URLCONF = 'enhydris.urls'

INSTALLED_APPS = (
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.staticfiles',
    'django.contrib.sites',
    'django.contrib.humanize',
    'django.contrib.gis',
    'django.contrib.flatpages',

    'rest_framework',

    'ajax_select',
    'captcha',
    'bootstrap3',

    'enhydris.hcore',
    'enhydris.api',
    'enhydris.permissions',

    # enhydris.hcore overrides some templates from django.contrib.admin; for
    # this reason, it must be listed in INSTALLED_APPS before
    # django.contrib.admin.
    'django.contrib.admin',
    # Registration templates it must be listed in INSTALLED_APPS after
    # 'enhydris.hcore' in order to overide django-default templates.
    'registration',
)

MIDDLEWARE_CLASSES = (
    'django.middleware.common.CommonMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.locale.LocaleMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.middleware.gzip.GZipMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.contrib.flatpages.middleware.FlatpageFallbackMiddleware',
)

APPEND_SLASH = True

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.core.context_processors.debug',
                'django.core.context_processors.i18n',
                'django.core.context_processors.media',
                'django.core.context_processors.static',
                'django.core.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
                'enhydris.hcore.context_processors.registration',
                'enhydris.hcore.context_processors.map',
            ],
        },
    },
]

AUTH_PROFILE_MODULE = 'hcore.UserProfile'
LOGIN_REDIRECT_URL = '/'

ATOMIC_REQUESTS = True
TEST_RUNNER = 'django.test.runner.DiscoverRunner'
USE_TZ = True

# Options for django-registration
ACCOUNT_ACTIVATION_DAYS = 7
REGISTRATION_OPEN = True

# Options for django-ajax-selects
AJAX_LOOKUP_CHANNELS = {
    'maintainers': dict(model='auth.User', search_field='username'),
}

# Default Enhydris settings
ENHYDRIS_FILTER_DEFAULT_COUNTRY = None
ENHYDRIS_FILTER_POLITICAL_SUBDIVISION1_NAME = None
ENHYDRIS_FILTER_POLITICAL_SUBDIVISION2_NAME = None
ENHYDRIS_USERS_CAN_ADD_CONTENT = False
ENHYDRIS_SITE_CONTENT_IS_FREE = False
ENHYDRIS_TSDATA_AVAILABLE_FOR_ANONYMOUS_USERS = False
ENHYDRIS_MIN_VIEWPORT_IN_DEGS = 0.04
ENHYDRIS_MAP_DEFAULT_VIEWPORT = (19.3, 34.75, 29.65, 41.8)
ENHYDRIS_TIMESERIES_DATA_DIR = 'timeseries_data'
ENHYDRIS_TS_GRAPH_BIG_STEP_DENOMINATOR = 200
ENHYDRIS_TS_GRAPH_FINE_STEP_DENOMINATOR = 50
ENHYDRIS_SITE_STATION_FILTER = {}
ENHYDRIS_DISPLAY_COPYRIGHT_INFO = False
ENHYDRIS_WGS84_NAME = 'WGS84'
ENHYDRIS_MAP_BASE_LAYERS = [
    r'OpenLayers.Layer.OSM.Mapnik("Open Street Map",'
    r'{isBaseLayer:true,attribution:'
    r'''"Map by <a href='http://www.openstreetmap.org/'>OSM</a>"})''',

    r'OpenLayers.Layer.OSM.CycleMap("Open Cycle Map",'
    r'{isBaseLayer: true, attribution:'
    r'''"Map by <a href='http://www.openstreetmap.org/'>OSM</a>"})''',
]
ENHYDRIS_MAP_BOUNDS = ((19.3, 34.75), (29.65, 41.8))
ENHYDRIS_MAP_MARKERS = {'0': 'images/drop_marker.png'}



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
import django.utils.timezone


class Migration(migrations.Migration):

    dependencies = [
        ('hcore', '0002_maintainers'),
    ]

    operations = [
        migrations.AlterField(
            model_name='eventtype',
            name='last_modified',
            field=models.DateTimeField(default=django.utils.timezone.now, null=True, editable=False),
        ),
        migrations.AlterField(
            model_name='filetype',
            name='last_modified',
            field=models.DateTimeField(default=django.utils.timezone.now, null=True, editable=False),
        ),
        migrations.AlterField(
            model_name='gentity',
            name='last_modified',
            field=models.DateTimeField(default=django.utils.timezone.now, null=True, editable=False),
        ),
        migrations.AlterField(
            model_name='gentityaltcode',
            name='last_modified',
            field=models.DateTimeField(default=django.utils.timezone.now, null=True, editable=False),
        ),
        migrations.AlterField(
            model_name='gentityaltcodetype',
            name='last_modified',
            field=models.DateTimeField(default=django.utils.timezone.now, null=True, editable=False),
        ),
        migrations.AlterField(
            model_name='gentityevent',
            name='last_modified',
            field=models.DateTimeField(default=django.utils.timezone.now, null=True, editable=False),
        ),
        migrations.AlterField(
            model_name='gentityfile',
            name='last_modified',
            field=models.DateTimeField(default=django.utils.timezone.now, null=True, editable=False),
        ),
        migrations.AlterField(
            model_name='gentitygenericdata',
            name='last_modified',
            field=models.DateTimeField(default=django.utils.timezone.now, null=True, editable=False),
        ),
        migrations.AlterField(
            model_name='gentitygenericdatatype',
            name='last_modified',
            field=models.DateTimeField(default=django.utils.timezone.now, null=True, editable=False),
        ),
        migrations.AlterField(
            model_name='instrument',
            name='last_modified',
            field=models.DateTimeField(default=django.utils.timezone.now, null=True, editable=False),
        ),
        migrations.AlterField(
            model_name='instrumenttype',
            name='last_modified',
            field=models.DateTimeField(default=django.utils.timezone.now, null=True, editable=False),
        ),
        migrations.AlterField(
            model_name='intervaltype',
            name='last_modified',
            field=models.DateTimeField(default=django.utils.timezone.now, null=True, editable=False),
        ),
        migrations.AlterField(
            model_name='lentity',
            name='last_modified',
            field=models.DateTimeField(default=django.utils.timezone.now, null=True, editable=False),
        ),
        migrations.AlterField(
            model_name='overseer',
            name='last_modified',
            field=models.DateTimeField(default=django.utils.timezone.now, null=True, editable=False),
        ),
        migrations.AlterField(
            model_name='stationtype',
            name='last_modified',
            field=models.DateTimeField(default=django.utils.timezone.now, null=True, editable=False),
        ),
        migrations.AlterField(
            model_name='timeseries',
            name='last_modified',
            field=models.DateTimeField(default=django.utils.timezone.now, null=True, editable=False),
        ),
        migrations.AlterField(
            model_name='timestep',
            name='last_modified',
            field=models.DateTimeField(default=django.utils.timezone.now, null=True, editable=False),
        ),
        migrations.AlterField(
            model_name='timezone',
            name='last_modified',
            field=models.DateTimeField(default=django.utils.timezone.now, null=True, editable=False),
        ),
        migrations.AlterField(
            model_name='unitofmeasurement',
            name='last_modified',
            field=models.DateTimeField(default=django.utils.timezone.now, null=True, editable=False),
        ),
        migrations.AlterField(
            model_name='variable',
            name='last_modified',
            field=models.DateTimeField(default=django.utils.timezone.now, null=True, editable=False),
        ),
    ]



"""
Hcore Widgets
"""

from django.utils.html import escape
from django.utils.safestring import mark_safe
from django.forms.utils import flatatt
from django.template.loader import render_to_string
import django.forms as forms

class SelectWithPop(forms.Select):
    """
    This widget creates a popup with a plus button on the right which
    opens a form for inline instance creation of foreign models
    """
    def __init__(self, model_name='', *args, **kwargs):
        """
        Overiden init method to populate gentities
        """
        self.model_name = model_name
        super(SelectWithPop, self).__init__(*args, **kwargs)

    def render(self, name, *args, **kwargs):
        if not self.model_name:
            self.model_name = name
        html = super(SelectWithPop, self).render(name, *args, **kwargs)
        popupplus = render_to_string("form/popplus.html",
                                                    {'field': self.model_name,
                                                     'orig_name': name})
        return html+popupplus


class ReadOnlyWidget(forms.Widget):
    """
    This is a widget for read only form fields.
    """
    def render(self, name, value, attrs):
        final_attrs = self.build_attrs(attrs, name=name)
        if hasattr(self, 'initial'):
            value = self.initial
            return mark_safe("<p %s>%s</p>" % (flatatt(final_attrs),
escape(value) or ''))

    def _has_changed(self, initial, data):
        return False

class ReadOnlyField(forms.Field):
    """
    This is a formfield similar to editable=False but actually displays the
    value of the field.
    """
    widget = ReadOnlyWidget
    def __init__(self, widget=None, label=None, initial=None, help_text=None):
        super(type(self), self).__init__(self, label=label, initial=initial,
            help_text=help_text, widget=widget)
        self.widget.initial = initial

    def clean(self, value):
        return self.widget.initial



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


def setup_group_names(apps, schema_editor):
    """Set the name of each group to be the slug with a capital initial."""
    SynopticGroup = apps.get_model("enhydris_synoptic", "SynopticGroup")
    for synoptic_group in SynopticGroup.objects.all():
        synoptic_group.name = synoptic_group.slug.title()
        synoptic_group.save()


class Migration(migrations.Migration):

    dependencies = [
        ('enhydris_synoptic', '0005_auto_20160421_1119'),
    ]

    operations = [
        migrations.RunPython(setup_group_names,
                             reverse_code=migrations.RunPython.noop),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('hcore', '0005_remove_is_active'),
    ]

    operations = [
        migrations.AlterField(
            model_name='timeseries',
            name='timestamp_offset_minutes',
            field=models.IntegerField(help_text='If unsure, set this to zero. It indicates the difference of what is shown from what is meant. For example, if for an hourly time series it is -5, then 2015-10-14 11:00 means the interval from 2015-10-14 09:55 to 2015-10-14 10:55. -1440 is common for daily time series.', null=True, blank=True),
        ),
        migrations.AlterField(
            model_name='timeseries',
            name='timestamp_offset_months',
            field=models.SmallIntegerField(help_text='If unsure, set this to 1 for monthly, 12 for annual, and zero otherwise.  For a monthly time series, an offset of -475 minutes and 1 month means that 2003-11-01 00:00 (normally shown as 2003-11) denotes the interval 2003-10-31 18:05 to 2003-11-30 18:05.', null=True, blank=True),
        ),
        migrations.AlterField(
            model_name='timeseries',
            name='timestamp_rounding_minutes',
            field=models.PositiveIntegerField(help_text='For an hourly time series whose timestamps end in :00, set this to zero; if they end in :12, set it to 12. For a ten-minute time series with timestamps ending in :12, :22, :32, etc., set it to 2.  For daily ending at 08:00, set it to 480. Leave empty if timestamps are irregular.', null=True, blank=True),
        ),
        migrations.AlterField(
            model_name='timeseries',
            name='timestamp_rounding_months',
            field=models.PositiveSmallIntegerField(help_text='Set this to zero, except for annual time series, indicating the difference from January; for example, set it to 9 if the timestamps use a hydrological year starting in October. Leave empty if timestamps are irregular.', null=True, blank=True),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models

import enhydris.hcore.models


class Migration(migrations.Migration):

    dependencies = [
        ('hcore', '0006_offset_help'),
    ]

    operations = [
        migrations.AddField(
            model_name='timeseries',
            name='datafile',
            field=models.FileField(storage=enhydris.hcore.models.TimeseriesStorage(), null=True, upload_to=b'', blank=True),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
import django.core.files.storage


class Migration(migrations.Migration):

    dependencies = [
        ('hcore', '0009_remove_ts_records'),
    ]

    operations = [
        migrations.AddField(
            model_name='timeseries',
            name='end_date',
            field=models.DateTimeField(null=True, blank=True),
        ),
        migrations.AddField(
            model_name='timeseries',
            name='start_date',
            field=models.DateTimeField(null=True, blank=True),
        ),
        migrations.AlterField(
            model_name='timeseries',
            name='datafile',
            field=models.FileField(storage=django.core.files.storage.FileSystemStorage(location=b'timeseries_data'), null=True, upload_to=b'', blank=True),
        ),
    ]



from django.conf import settings
from django.shortcuts import get_object_or_404

from rest_framework import permissions

from enhydris.hcore import models


class CanEditOrReadOnly(permissions.BasePermission):
    """
    Custom permission to only allow allowed users to edit an object.
    """

    def has_object_permission(self, request, view, obj):
        if request.method in permissions.SAFE_METHODS:
            return True
        if isinstance(obj, models.Timeseries):
            id = obj.gentity.id
        elif isinstance(obj, models.Gentity):
            id = obj.id
        else:
            return False
        station = get_object_or_404(models.Station, id=id)
        return hasattr(request.user, 'has_row_perm') \
            and request.user.has_row_perm(station, 'edit')


class CanCreateStation(permissions.BasePermission):

    def has_permission(self, request, view):
        if request.method in permissions.SAFE_METHODS:
            return True
        if request.method != 'POST' or not request.user.is_authenticated():
            return False
        if settings.ENHYDRIS_USERS_CAN_ADD_CONTENT or request.user.has_perm(
                'hcore.add_station'):
            return True
        return False



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import connection, migrations
from django.db.utils import ProgrammingError


def delete_database_functions(apps, schema_editor):
    cursor = connection.cursor()
    cursor.execute(
        "DROP FUNCTION IF EXISTS timeseries_start_date(aid INTEGER)")
    cursor.execute(
        "DROP FUNCTION IF EXISTS timeseries_end_date(aid INTEGER)")


def undelete_database_functions(apps, schema_editor):
    cursor = connection.cursor()
    cursor.execute("""
        CREATE FUNCTION timeseries_start_date(aid INTEGER)
        RETURNS timestamp AS $$
        DECLARE
            retvalue timestamp;
        BEGIN
            SELECT INTO retvalue
                to_timestamp(substring(CASE WHEN top='' OR top IS
                                        NULL THEN bottom ELSE top END
                    from '^(.*?),'),
                    'YYYY-MM-DD HH24:MI')::timestamp
            FROM ts_records
            WHERE id=aid;
            RETURN retvalue;
        END
        $$ LANGUAGE plpgsql""")
    cursor.execute("""
        CREATE FUNCTION timeseries_end_date(aid INTEGER)
        RETURNS timestamp AS $$
        DECLARE
            retvalue timestamp;
        BEGIN
            SELECT INTO retvalue to_timestamp(
                    substring(bottom from E'\n([^,]*?),[^\n]*?\n?$'),
                    'YYYY-MM-DD HH24:MI')::timestamp
            FROM ts_records
            WHERE id=aid;
            RETURN retvalue;
        END
        $$ LANGUAGE plpgsql""")


class Migration(migrations.Migration):

    dependencies = [
        ('hcore', '0008_move_data_to_datafile'),
    ]

    operations = [
        migrations.RunPython(delete_database_functions,
                             reverse_code=undelete_database_functions),
        migrations.DeleteModel(
            name='TsRecords',
        ),
    ]



from django.utils.translation import ugettext as _


def check_latitude(latitude):
    if latitude > 90 or latitude < -90:
        raise Exception(_("%f is not a valid latitude") % (latitude,))


def check_longitude(longitude):
    if longitude > 180 or longitude < -180:
        raise Exception(_("%f is not a valid longitude") % (longitude,))


def check_altitude(altitude):
    if altitude > 8850 or altitude < -422:
        raise Exception(_("%f is not a valid altitude") % (altitude,))



from __future__ import absolute_import, unicode_literals

from datetime import timedelta
import os
import platform
from six import BytesIO, StringIO
import sys
import time

from django.conf import settings
from django.template.loader import render_to_string

import matplotlib; matplotlib.use('AGG')  # NOQA
from matplotlib.dates import DateFormatter, DayLocator, HourLocator
import matplotlib.pyplot as plt
import pandas as pd

from enhydris_synoptic.celery import app
from enhydris_synoptic import models
from enhydris_synoptic.models import SynopticGroup, SynopticGroupStation


def write_output_to_file(relative_filename, s, binary=False):
    """Write string (or bytes) to a file.

       The resulting output file name is the concatenation of
       ENHYDRIS_SYNOPTIC_ROOT plus relative_filename.
    """
    # Determine the output file name
    output_file = os.path.join(settings.ENHYDRIS_SYNOPTIC_ROOT,
                               relative_filename)

    # Make sure the directory exists
    dirname = os.path.dirname(output_file)
    if not os.path.exists(dirname):
        os.makedirs(dirname)

    # Write result to temporary file
    temporary_file = output_file + '.1'
    s = s if binary else s.encode('utf8')
    mode = 'wb' if binary else 'w'
    with open(temporary_file, mode) as f:
        f.write(s)

    # Replace final file, atomically if possible
    if sys.version_info[0] >= '3':
        os.replace(temporary_file, output_file)
    elif platform.system() == 'Windows':
        if os.path.exists(output_file):
            removed = False
            for i in range(0, 100):
                try:
                    os.remove(output_file)
                    removed = True
                    break
                except:
                    time.sleep(0.001)
            if not removed:
                os.remove(output_file)  # This time raise the exception
        os.rename(temporary_file, output_file)
    else:
        os.rename(temporary_file, output_file)


def get_last_common_date(synoptic_timeseries):
    # We don't actually return the last common date, which would be
    # difficult; instead, we return the minimum of the last dates of the
    # timeseries, which will usually be the last common date. station is
    # an enhydris_synoptic.models.Station object.
    result = None
    for asynts in synoptic_timeseries:
        end_date = asynts.timeseries.end_date
        if end_date and ((not result) or (end_date < result)):
            result = end_date
    return result


def get_timeseries_for_synoptic_group_station(sgroupstation):
    """Return list with synoptic timeseries.

    For each synoptic timeseries in the given synoptic group station,
    grab the last 144 records preceding the last common date, attach these
    records to the synoptic timeseries object, and attach these time series
    to the sgroupstation object as the synoptic_timeseries attribute.
    """
    synoptic_timeseries = list(models.SynopticTimeseries.objects.filter(
        synoptic_group_station=sgroupstation))
    sgroupstation.last_common_date = get_last_common_date(synoptic_timeseries)
    for asynts in synoptic_timeseries:
        # Get the time series records
        tsrecords = asynts.timeseries.get_all_data()

        # Keep only the 144 preceding last_common_date
        bounding_dates = tsrecords.bounding_dates()
        if bounding_dates:
            start_date, end_date = bounding_dates
            tsrecords.delete_items(
                sgroupstation.last_common_date + timedelta(minutes=1),
                end_date)
            tsrecords.delete_items(
                start_date, sgroupstation.last_common_date - timedelta(days=1))

        # Attach records to the object
        asynts.tsrecords = tsrecords
    sgroupstation.synoptic_timeseries = synoptic_timeseries


def add_synoptic_group_station_context(synoptic_group_station):
    for asynts in synoptic_group_station.synoptic_timeseries:
        synoptic_group_station.error = False
        try:
            asynts.value = asynts.tsrecords[
                synoptic_group_station.last_common_date]
        except KeyError:
            synoptic_group_station.error = True
            continue


def render_synoptic_group(synoptic_group, all_sgroupstations):
    subset = [x
              for x in all_sgroupstations
              if x.synoptic_group.id == synoptic_group.id]
    output = render_to_string(
        'synopticgroup.html',
        context={'object': synoptic_group,
                 'synoptic_group_stations': subset,
                 })
    filename = os.path.join(synoptic_group.slug, 'index.html')
    write_output_to_file(filename, output)


def render_synoptic_station(sgroupstation):
    output = render_to_string('synopticgroupstation.html',
                              context={'object': sgroupstation})
    filename = os.path.join(sgroupstation.synoptic_group.slug,
                            'station', str(sgroupstation.id), 'index.html')
    write_output_to_file(filename, output)


def get_color(i):
    """Return the color to be used for line with sequence i

    The first line to be drawn uses red, so self.get_color(0)='red';
    the second one uses green, so self.get_color(1)='green'; and there are
    also one or two more colors. We assume the user will not attempt to
    group more than a few time series together. However, if this happens,
    we recycle the colors starting from red again, to make sure we don't
    have an index error or anything.
    """
    colors = ['red', 'green', 'blue', 'magenta']
    return colors[i % len(colors)]


def render_chart(a_synoptic_timeseries, all_synoptic_timeseries):
    # Get all synoptic time series to put in the chart; i.e. the
    # requested one plus the ones that are grouped with it.
    synoptic_timeseries = [
        x for x in all_synoptic_timeseries
        if (x.id == a_synoptic_timeseries.id)
        or (x.group_with and (x.group_with.id == a_synoptic_timeseries.id))]

    # Read the last 144 records of each time series into a pandas object
    for t in synoptic_timeseries:
        buff = StringIO()
        t.tsrecords.write(buff)
        buff.seek(0)
        t.pandas_object = pd.read_csv(buff, parse_dates=[0],
                                      usecols=[0, 1], index_col=0,
                                      header=None)

    # Reorder them so that the greatest is at the top
    synoptic_timeseries.sort(key=lambda x: float(x.pandas_object.sum()),
                             reverse=True)

    # Setup plot
    fig = plt.figure()
    fig.set_dpi(100)
    fig.set_size_inches(3.2, 2)
    fig.subplots_adjust(left=0.10, right=0.99, bottom=0.15, top=0.97)
    matplotlib.rcParams.update({'font.size': 7})
    ax = fig.add_subplot(1, 1, 1)

    # Draw lines. We use matplotlib's plot() instead of pandas's wrapper,
    # because otherwise there is trouble modifying the x axis labels
    # (see http://stackoverflow.com/questions/12945971/).
    for i, s in enumerate(synoptic_timeseries):
        xdata = s.pandas_object.index.to_pydatetime()
        ydata = s.pandas_object.values.transpose()[0]
        ax.plot(xdata, ydata, color=get_color(i),
                label=s.get_subtitle())
        if i == 0:
            # We will later need the data of the first time series, in
            # order to fill the chart
            gydata = ydata

    # Change plot limits as needed
    xmin, xmax, ymin, ymax = ax.axis()
    if a_synoptic_timeseries.default_chart_min:
        ymin = min(a_synoptic_timeseries.default_chart_min, ymin)
    if a_synoptic_timeseries.default_chart_max:
        ymax = max(a_synoptic_timeseries.default_chart_max, ymax)
    ax.set_ylim([ymin, ymax])

    # Fill
    ax.fill_between(xdata, gydata, ymin, color='#ffff00')

    # X ticks and labels
    ax.xaxis.set_minor_locator(HourLocator(byhour=range(0, 24, 3)))
    ax.xaxis.set_minor_formatter(DateFormatter('%H:%M'))
    ax.xaxis.set_major_locator(DayLocator())
    ax.xaxis.set_major_formatter(
        DateFormatter('\n    %Y-%m-%d $\\rightarrow$'))

    # Gridlines and legend
    ax.grid(b=True, which='both', color='b', linestyle=':')
    if len(synoptic_timeseries) > 1:
        ax.legend()

    # Create and save plot
    f = BytesIO()
    fig.savefig(f)
    plt.close(fig)  # Release some memory
    filename = os.path.join('chart', str(a_synoptic_timeseries.id) + '.png')
    write_output_to_file(filename, f.getvalue(), binary=True)
    f.close()

    # If unit testing, write the data to a file.
    if hasattr(settings, 'TEST_MATPLOTLIB') and settings.TEST_MATPLOTLIB:
        filename = os.path.join('chart',
                                str(a_synoptic_timeseries.id) + '.dat')
        data = [repr(line.get_xydata()).replace('\n', ' ')
                for line in ax.lines]
        write_output_to_file(filename, '(' + ', '.join(data) + ')')


@app.task
def create_static_files():
    """Create static html files for all enhydris-synoptic."""
    # Create a list of all synoptic group stations
    all_sgroupstations = []
    for sgroupstation in SynopticGroupStation.objects.all():
        get_timeseries_for_synoptic_group_station(sgroupstation)
        add_synoptic_group_station_context(sgroupstation)
        all_sgroupstations.append(sgroupstation)

    # For each station, create its reports and its charts
    for sgroupstation in all_sgroupstations:
        render_synoptic_station(sgroupstation)
        for t in sgroupstation.synoptic_timeseries:
            render_chart(t, sgroupstation.synoptic_timeseries)

    # Finally create the reports for each synoptic group
    for sgroup in SynopticGroup.objects.all():
        render_synoptic_group(sgroup, all_sgroupstations)



from importlib import import_module
import os

from django.core.management.base import CommandError
from django.core.management.templates import TemplateCommand
from django.utils.crypto import get_random_string

import enhydris


class Command(TemplateCommand):
    help = ("Creates an Enhydris project directory structure for the given "
            "project name in the current directory or optionally in the given "
            "directory.")

    def add_arguments(self, parser):
        parser.add_argument('name',
                            help='Name of the application or project.')
        parser.add_argument('directory', nargs='?',
                            help='Optional destination directory')

    def handle(self, *args, **options):
        instance_name, target = options.pop('name'), options.pop('directory')
        self.validate_name(instance_name, 'project')

        # Check that the instance_name cannot be imported
        try:
            import_module(instance_name)
        except ImportError:
            pass
        else:
            raise CommandError("{} conflicts with the name of an existing "
                               "Python module and cannot be used as an "
                               "instance name. Please try another name."
                               .format(instance_name))

        # Create a random SECRET_KEY to put in the main settings
        chars = 'abcdefghijklmnopqrstuvwxyz0123456789!@#$%^&*(-_=+)'
        options['secret_key'] = get_random_string(50, chars)

        options['template'] = os.path.join(enhydris.__path__[0], 'conf')
        options['extensions'] = ['py']
        options['files'] = []
        options['enhydris_version'] = enhydris.__version__
        version_items = enhydris.__version__.split('.')
        if len(version_items) > 1:
            enhydris_docs_version = '{}.{}'.format(*version_items[:2])
        else:
            enhydris_docs_version = 'latest'
        options['enhydris_docs_version'] = enhydris_docs_version

        super(Command, self).handle('project', instance_name, target,
                                    **options)



# -*- coding: utf-8 -*-
# UTF8 Encoded
from django import forms
from django.contrib import admin
from django.utils.translation import ugettext_lazy as _
from enhydris.hcore.models import *
from enhydris.hcore.forms import *

##########################################

class LentityAdmin(admin.ModelAdmin):
    pass

class OrganizationAdmin(LentityAdmin):
    list_display = [f.name for f in Organization._meta.fields]
    list_display_links = [f.name for f in Organization._meta.fields]

class PersonAdmin(LentityAdmin):
    list_display = [f.name for f in Person._meta.fields]
    list_display_links = [f.name for f in Person._meta.fields]

class GentityAdmin(admin.ModelAdmin):
    pass

class GareaAdmin(GentityAdmin):
    pass

class PoliticalDivisionAdmin(GareaAdmin):
    list_display = [f.name for f in PoliticalDivision._meta.fields]
    list_display_links = [f.name for f in PoliticalDivision._meta.fields]

class WaterDivisionAdmin(GareaAdmin):
    # TODO: Fill it when the model is ready
    pass

class WaterBasinAdmin(GareaAdmin):
    list_display = [f.name for f in WaterBasin._meta.fields]
    list_display_links = [f.name for f in WaterBasin._meta.fields]

class GentityFileAdmin(admin.ModelAdmin):
    list_display = [f.name for f in GentityFile._meta.fields]

class GentityGenericDataAdmin(admin.ModelAdmin):
    list_display = [f.name for f in GentityGenericData._meta.fields]

class GentityAltCodeTypeAdmin(admin.ModelAdmin):
    list_display = [f.name for f in GentityAltCodeType._meta.fields]

class FileTypeAdmin(admin.ModelAdmin):
    list_display = ('id', 'mime_type', 'descr',)

class GentityGenericDataTypeAdmin(admin.ModelAdmin):
    list_display = ('id', 'descr', 'file_extension', )

class EventTypeAdmin(admin.ModelAdmin):
    list_display = ('id', 'descr',)

class StationTypeAdmin(admin.ModelAdmin):
    list_display = ('id', 'descr',)

class InstrumentTypeAdmin(admin.ModelAdmin):
    list_display = [f.name for f in InstrumentType._meta.fields]

class VariableAdmin(admin.ModelAdmin):
    list_display = [f.name for f in Variable._meta.fields]

class IntervalTypeAdmin(admin.ModelAdmin):
    list_display = [f.name for f in IntervalType._meta.fields]

class UnitOfMeasurementAdmin(admin.ModelAdmin):
    list_display = [f.name for f in UnitOfMeasurement._meta.fields]

class TimeStepAdmin(admin.ModelAdmin):
    form = TimeStepForm
    list_display = [f.name for f in TimeStep._meta.fields]

class TimeZoneAdmin(admin.ModelAdmin):
    list_display = [f.name for f in TimeZone._meta.fields]

class UserProfileAdmin(admin.ModelAdmin):
    pass

admin.site.register(Lentity, LentityAdmin)
admin.site.register(Organization, OrganizationAdmin)
admin.site.register(Person, PersonAdmin)

admin.site.register(Gentity, GentityAdmin)
admin.site.register(Garea, GareaAdmin)
admin.site.register(PoliticalDivision, PoliticalDivisionAdmin)
admin.site.register(WaterDivision, WaterDivisionAdmin)
admin.site.register(WaterBasin, WaterBasinAdmin)

admin.site.register(GentityAltCodeType, GentityAltCodeTypeAdmin)
admin.site.register(GentityFile, GentityFileAdmin)
admin.site.register(GentityGenericData, GentityGenericDataAdmin)
admin.site.register(FileType, FileTypeAdmin)
admin.site.register(GentityGenericDataType, GentityGenericDataTypeAdmin)
admin.site.register(EventType, EventTypeAdmin)
admin.site.register(StationType, StationTypeAdmin)
admin.site.register(InstrumentType, InstrumentTypeAdmin)
admin.site.register(Variable, VariableAdmin)
admin.site.register(UnitOfMeasurement, UnitOfMeasurementAdmin)
admin.site.register(TimeStep, TimeStepAdmin)
admin.site.register(TimeZone, TimeZoneAdmin)
admin.site.register(IntervalType, IntervalTypeAdmin)
admin.site.register(UserProfile, UserProfileAdmin)

##########################################

class GentityAltCodeInline(admin.TabularInline):
    model = GentityAltCode
    extra = 1

class GentityFileInline(admin.TabularInline):
    model = GentityFile
    extra = 1

class GentityGenericDataInline(admin.TabularInline):
    model = GentityGenericData
    extra = 1
 
class GentityEventInline(admin.TabularInline):
    model = GentityEvent
    extra = 1

# M2M through Overseer
class OverseerInline(admin.TabularInline):
    model = Overseer
    extra = 1

class InstrumentInline(admin.TabularInline):
    model = Instrument
    extra = 1

class TimeseriesInline(admin.TabularInline):
    model = Timeseries
    extra = 1


class StationAdminForm(forms.ModelForm):

    class Meta:
        model = Station
        exclude = []
    def clean_altitude(self):
        value = self.cleaned_data['altitude']
        if not value == None and (value > 8850 or value < -422):
            raise forms.ValidationError(_("%f is not a valid altitude") %
            (value,))
        return self.cleaned_data['altitude']


class StationAdmin(admin.ModelAdmin):
    form = StationAdminForm
    # ChangeList format
    list_display = ('id', 'name', 'short_name', 'remarks', 'water_basin',
        'water_division', 'political_division', 'srid',
        'approximate', 'altitude', 'asrid', 'owner', 'is_automatic',
        'start_date', 'end_date','show_overseers')
    list_filter = ('is_automatic', 'end_date',)
    date_hierarchy = 'start_date'

    # Inlines for m2m relationships
    inlines = (GentityAltCodeInline, GentityFileInline, 
                GentityGenericDataInline, GentityEventInline,
                OverseerInline, InstrumentInline, TimeseriesInline,)

admin.site.register(Station, StationAdmin)

class InstrumentAdmin(admin.ModelAdmin):
    # ChangeList format
    list_display = ('id', 'name', 'remarks', 'station', 'type', 'manufacturer',
        'model', 'start_date', 'end_date' )
    list_filter = ('start_date', 'end_date',)

admin.site.register(Instrument, InstrumentAdmin)

class TimeseriesAdmin(admin.ModelAdmin):
    form = TimeseriesDataForm
    # ChangeList format
    list_display = ('id', 'name', 'remarks', 'gentity', 'variable', 
        'unit_of_measurement', 'precision', 'time_zone', 'instrument',
        'time_step', 'timestamp_rounding_minutes', 'timestamp_rounding_months', 
        'interval_type', 'timestamp_offset_minutes', 'timestamp_offset_months')
    list_filter = ('unit_of_measurement','precision', 'time_zone', 
                    'timestamp_rounding_minutes', 'timestamp_rounding_months',
                    'timestamp_offset_minutes', 'timestamp_offset_months')


admin.site.register(Timeseries, TimeseriesAdmin)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('enhydris_synoptic', '0003_auto_20151021_1041'),
    ]

    operations = [
        migrations.AddField(
            model_name='synoptictimeseries',
            name='group_with',
            field=models.ForeignKey(blank=True, to='enhydris_synoptic.SynopticTimeseries', help_text='Specify this field if you want to group this time series with another in the same chart and in the report.', null=True),
        ),
        migrations.AddField(
            model_name='synoptictimeseries',
            name='subtitle',
            field=models.CharField(help_text='If time series are grouped, this is shows in the legend of the chart and in the report, in brackets.', max_length=50, blank=True),
        ),
        migrations.AddField(
            model_name='synoptictimeseries',
            name='title',
            field=models.CharField(help_text='Used as the chart title and as the time series title in the report. Leave empty to use the time series name.', max_length=50, blank=True),
        ),
    ]
